//---------------------------------------------------------------------------//
//
//	Graphics - Renderer
//
//	Date : 2008/04/07
//
//---------------------------------------------------------------------------//

#include "graphics_std.h"
#include "renderer.h"

CRenderer::CRenderer() 
: m_pD3D(NULL), m_pDevice(NULL), m_pBackBuffer(NULL), m_pSprite(NULL), m_pFont(NULL)
{
	m_dwNumRenderedPrimitive = 0;
	m_dwNumCallDrawPrimitive = 0;
	m_bDrawTextureMode = false;
	m_bFogEnable = false;
	m_bLighting = false;
	m_bSpecularLighting = false;
	m_bAlphaTestEnable = false;
	m_btAlphaTestRef = 0;
	m_bZWriteEnable = false;
	m_bZEnable = true;
	m_bWireframe = false;
	memset(&m_bLightEnable, 0, sizeof(bool)*MAX_ACTIVE_LIGHT);

}

CRenderer::~CRenderer()
{
	Release();
}

//!  ʱȭ Ѵ.
/*!
	@param hWnd  ڵ
	@param nWidth   ũ
	@param nHeight   ũ
	@param bWindowMode   
	@return  ʱȭ 
*/
bool CRenderer::Init(HWND hWnd, int nWidth, int nHeight, bool bFullScreen)
{
	HRESULT	hr;

	m_pD3D = Direct3DCreate9(D3D_SDK_VERSION);
	if (m_pD3D == NULL)
	{
		MessageBox(hWnd, _T("Direct3D ü ϴµ Ͽϴ."), _T("Error"), MB_OK | MB_ICONWARNING);
		MessageBox(hWnd, _T("Direct9.0c ̻ DirectX  ʿմϴ."), _T("Error"), MB_OK | MB_ICONWARNING);
		return false;
	}

	hr = m_pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &m_d3ddm);
	if (FAILED(hr))
	{
		MessageBox(hWnd, _T("÷ 带 ü ϴ."), _T("Error"), MB_OK | MB_ICONWARNING);
		return false;
	}

	memset(&m_d3dpp, 0, sizeof(D3DPRESENT_PARAMETERS));

	m_d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	m_d3dpp.hDeviceWindow = hWnd;
	m_d3dpp.BackBufferCount = 1;
	m_d3dpp.EnableAutoDepthStencil = TRUE;
	m_d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
	m_d3dpp.Flags = D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;
	m_d3dpp.BackBufferWidth = nWidth;
	m_d3dpp.BackBufferHeight = nHeight;

	m_bFullScreen = bFullScreen;
	if (m_bFullScreen) // fullscreen mode
	{
		m_d3dpp.Windowed = false;
		m_d3dpp.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT; //   ֻ .
		m_d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT; // ֻ   Ѵ.
		m_d3dpp.BackBufferFormat = D3DFMT_X8R8G8B8;
	}
	else // window mode
	{
		m_d3dpp.Windowed = true;
		m_d3dpp.FullScreen_RefreshRateInHz = 0;
		m_d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
		m_d3dpp.BackBufferFormat = m_d3ddm.Format;
	}

	hr = m_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING, &m_d3dpp, &m_pDevice);
	if (FAILED(hr))
	{
		hr = m_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_SOFTWARE_VERTEXPROCESSING, &m_d3dpp, &m_pDevice);
		if (FAILED(hr))
		{	
			ASSERT_MSG(0, _T("creating direct3d device is failed.!!"));
			MessageBox(hWnd, _T("D3D ġ ϴµ ߽ϴ."), _T("Error"), MB_OK | MB_ICONWARNING);
			return false;
		}
	}

	hr = m_pDevice->GetBackBuffer(0, 0, D3DBACKBUFFER_TYPE_MONO, &m_pBackBuffer);
	if (FAILED(hr))
	{
		ASSERT_MSG(0, _T("GetBackBuffer() is failed.!!"));
		return false;
	}

	hr = D3DXCreateSprite(m_pDevice, &m_pSprite);
	if (FAILED(hr))
	{
		ASSERT_MSG(0, _T("Initializing sprite interface is failed.!!"));
		return false;
	}

	SetRenderer(this);

	return true;
}

//!   Ѵ. Ҹڿ ڵ ȣȴ.
void CRenderer::Release()
{
	SAFE_RELEASE(m_pBackBuffer);
	SAFE_RELEASE(m_pSprite);
	SAFE_RELEASE(m_pDevice);
	SAFE_RELEASE(m_pD3D);
}

void CRenderer::ClearBuffer(bool bClearBackBuffer, DWORD dwBackBufferColor, bool bClearDepthBuffer)
{
	DWORD dwFlags = 0;

	if (bClearBackBuffer)
	{
		dwFlags |= D3DCLEAR_TARGET;
	}
	if (bClearDepthBuffer)
	{
		dwFlags |= D3DCLEAR_ZBUFFER;
	}

	if (dwFlags != 0)
	{
        m_pDevice->Clear(0, NULL, dwFlags, dwBackBufferColor, 1.0f, 0);
    }
}

void CRenderer::BeginScene()
{
	m_pDevice->BeginScene();
}

void CRenderer::EndScene()
{
	m_pDevice->EndScene();
}

void CRenderer::Present(HWND hWnd)
{
	m_pDevice->Present(NULL, NULL, hWnd, NULL);
	m_dwNumRenderedPrimitive = 0;
	m_dwNumCallDrawPrimitive = 0;
}

void CRenderer::BeginDrawTexture(int nWidth, int nHeight)
{
	ASSERT(m_bDrawTextureMode == false);

	m_pSprite->Begin(D3DXSPRITE_ALPHABLEND | D3DXSPRITE_SORT_DEPTH_BACKTOFRONT);
	m_bDrawTextureMode = true;

	if (nWidth != 0 && nHeight != 0)
	{
		//  Ϸ Ŀ  κ  Ѵ.
		float ortho_11 = 1.0f / (float)(nWidth/2);
		float ortho_22 = -1.0f / (float)(nHeight/2);
	
		matrix4 matOrtho(ortho_11,     0.0f,    0.0f,    0.0f,
							 0.0f, ortho_22,    0.0f,    0.0f,
							 0.0f,	   0.0f,    0.0f,    0.0f,
							-1.0f,     1.0f,    0.0f,    1.0f);
		SetProj(matOrtho);
	}
}

void CRenderer::DrawTexture(CTexture* pTexture, int x, int y, DWORD dwColor)
{
	ASSERT(m_bDrawTextureMode == true && pTexture);
	m_pSprite->Draw(pTexture->GetTexture(), NULL, NULL, &vector3((float)x+0.5f, (float)y+0.5f, 0.0f), dwColor);
	//m_pSprite->Flush();
	m_dwNumRenderedPrimitive += 2;
}

void CRenderer::DrawTextureEx(CTexture* pTexture, int sx, int sy, int width, int height, const vector3* pvecPos, DWORD dwColor, const vector3* pvecCenter, const matrix4* pmatTransform)
{
	ASSERT(m_bDrawTextureMode == true && pTexture);

	if (pmatTransform)
	{
		m_pSprite->SetTransform(pmatTransform);
	}

	RECT SrcRect;
	SrcRect.left = sx;
	SrcRect.right = sx+width;
	SrcRect.top = sy;
	SrcRect.bottom = sy+height;
	m_pSprite->Draw(pTexture->GetTexture(), &SrcRect, pvecCenter, pvecPos, dwColor);
	//m_pSprite->Flush();
	m_dwNumRenderedPrimitive += 2;

	if (pmatTransform)
	{
		matrix4 matIdentity;
		mat4_identity(matIdentity);
		m_pSprite->SetTransform(&matIdentity);
	}
}

void CRenderer::DrawTextureEx(CTexture* pTexture, int sx, int sy, int width, int height, float dx, float dy, vector3& vecCenter, vector3& vecScale, DWORD dwColor, float fRadian)
{
	ASSERT(m_bDrawTextureMode == true && pTexture);

	matrix4 matTransform;
	mat4_identity(matTransform);
	matTransform._11 = vecScale.x;
	matTransform._22 = vecScale.y;
	matTransform._33 = vecScale.z;
	matrix4 matRot;
	mat4_rot_z(matRot, fRadian);
	mat4_multiply(matTransform, matTransform, matRot);

	matTransform._41 = dx;
	matTransform._42 = dy;
	m_pSprite->SetTransform(&matTransform);
	
	RECT SrcRect;
	SrcRect.left = sx;
	SrcRect.right = sx+width;
	SrcRect.top = sy;
	SrcRect.bottom = sy+height;
	m_pSprite->Draw(pTexture->GetTexture(), &SrcRect, &vecCenter, NULL, dwColor);
	m_dwNumRenderedPrimitive += 2;

	matrix4 matIdentity;
	mat4_identity(matIdentity);
	m_pSprite->SetTransform(&matIdentity);
}

void CRenderer::DrawTextureEx(CTexture* pTexture, int sx, int sy, int width, int height,
							  float dx, float dy, float cx, float cy, DWORD dwColor, 
							  float fScale, float fRadian, bool bFlipVert, bool bFlipHorz)
{
	ASSERT(m_bDrawTextureMode == true && pTexture);
	matrix4 matTransform;
	mat4_identity(matTransform);
	matTransform._11 = fScale;
	matTransform._22 = fScale;
	matTransform._33 = fScale;
	matrix4 matRot;
	mat4_rot_z(matRot, fRadian);
	mat4_multiply(matTransform, matTransform, matRot);
	if (bFlipVert || bFlipHorz)
	{
		matrix4 matFlip;
		mat4_identity(matFlip);
		if (bFlipVert)
			matFlip._11 = -1.0f;
		if (bFlipHorz)
			matFlip._22 = -1.0f;

		mat4_multiply(matTransform, matTransform, matFlip);
	}
	matTransform._41 = dx;
	matTransform._42 = dy;
	m_pSprite->SetTransform(&matTransform);
	vector3 vecCenter(cx, cy, 0.0f);
	RECT SrcRect;
	SrcRect.left = sx;
	SrcRect.right = sx+width;
	SrcRect.top = sy;
	SrcRect.bottom = sy+height;
	m_pSprite->Draw(pTexture->GetTexture(), &SrcRect, &vecCenter, NULL, dwColor);
	m_dwNumRenderedPrimitive += 2;
	matrix4 matIdentity;
	mat4_identity(matIdentity);
	m_pSprite->SetTransform(&matIdentity);
}

void CRenderer::FlushDrawTexture()
{
	ASSERT(m_bDrawTextureMode == true);
	m_pSprite->Flush();
}

void CRenderer::EndDrawTexture()
{
	ASSERT(m_bDrawTextureMode == true);

	m_pSprite->End();
	m_bDrawTextureMode = false;
}

void CRenderer::DrawTextDC(int x, int y, LPCTSTR cszText, COLORREF Color)
{
	HDC hDC = NULL;
	m_pBackBuffer->GetDC(&hDC);

	if (hDC != NULL)
	{
		SetBkMode(hDC, TRANSPARENT);
		SetTextColor(hDC, Color);
		TextOut(hDC, x, y, cszText, (int)_tcslen(cszText));
		m_pBackBuffer->ReleaseDC(hDC);
	}
}